/*
 * Copyright (C) 1997-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file command.c Client authentication and server commands. */

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <dirent.h>
#include <openssl/rc4.h>

#include "para.h"
#include "error.h"
#include "server.cmdline.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "vss.h"
#include "list.h"
#include "send.h"
#include "rc4.h"
#include "net.h"
#include "daemon.h"
#include "fd.h"
#include "ipc.h"
#include "user_list.h"
#include "server_command_list.h"
#include "afs_command_list.h"

/** Commands including options must be shorter than this. */
#define MAX_COMMAND_LEN 32768

static RC4_KEY rc4_recv_key;
static RC4_KEY rc4_send_key;
static unsigned char rc4_buf[2 * RC4_KEY_LEN];

extern int mmd_mutex;
extern struct misc_meta_data *mmd;
extern struct sender senders[];

static void dummy(int s)
{
	/*
	 * At least on Solaris, SIGUSR1 is one-shot, i.e. the signal action is
	 * restored to the default state once the signal handler has been
	 * called.
	 */
	if (s == SIGUSR1)
		signal(SIGUSR1, dummy);
}

static void mmd_dup(struct misc_meta_data *new_mmd)
{
	mutex_lock(mmd_mutex);
	*new_mmd = *mmd;
	mutex_unlock(mmd_mutex);
}

/*
 * Compute human readable string containing vss status for given integer value.
 *
 * We don't want to use vss_playing() and friends here because we take a
 * snapshot of the mmd struct and use the copy for computing the state of the
 * vss. If the real data were used, we would take the mmd lock for a rather
 * long time or risk to get an inconsistent view.
 */
static char *vss_status_tohuman(unsigned int flags)
{
	if (flags & VSS_PLAYING)
		return para_strdup("playing");
	if (flags & VSS_NEXT)
		return para_strdup("stopped");
	return para_strdup("paused");
}

/*
 * return human readable permission string. Never returns NULL.
 */
static char *cmd_perms_itohuman(unsigned int perms)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = perms & AFS_READ? 'a' : '-';
	msg[1] = perms & AFS_WRITE? 'A' : '-';
	msg[2] = perms & VSS_READ? 'v' : '-';
	msg[3] = perms & VSS_WRITE? 'V' : '-';
	msg[4] = '\0';
	return msg;
}

/*
 * Never returns NULL.
 */
static char *vss_get_status_flags(unsigned int flags)
{
	char *msg = para_malloc(5 * sizeof(char));

	msg[0] = (flags & VSS_PLAYING)? 'P' : '_';
	msg[1] = (flags & VSS_NOMORE)? 'O' : '_';
	msg[2] = (flags & VSS_NEXT)? 'N' : '_';
	msg[3] = (flags & VSS_REPOS)? 'R' : '_';
	msg[4] = '\0';
	return msg;
}

static char *get_status(struct misc_meta_data *nmmd)
{
	char *ret, mtime[30] = "";
	char *status, *flags; /* vss status info */
	char *ut = uptime_str();
	long offset = (nmmd->offset + 500) / 1000;
	struct timeval now;
	struct tm mtime_tm;

	/* report real status */
	status = vss_status_tohuman(nmmd->vss_status_flags);
	flags = vss_get_status_flags(nmmd->vss_status_flags);
	if (nmmd->size) { /* parent currently has an audio file open */
		localtime_r(&nmmd->mtime, &mtime_tm);
		strftime(mtime, 29, "%b %d %Y", &mtime_tm);
	}
	gettimeofday(&now, NULL);
	ret = make_message(
		"%s: %zu\n" /* file size */
		"%s: %s\n" /* mtime */
		"%s: %s\n" /* status */
		"%s: %s\n" /* status flags */
		"%s: %li\n" /* offset */
		"%s: %s\n" /* afs mode */
		"%s: %lu.%lu\n" /* stream start */
		"%s: %lu.%lu\n" /* current server time */
		"%s", /* afs status info */
		status_item_list[SI_FILE_SIZE], nmmd->size / 1024,
		status_item_list[SI_MTIME], mtime,
		status_item_list[SI_STATUS], status,
		status_item_list[SI_STATUS_FLAGS], flags,

		status_item_list[SI_OFFSET], offset,
		status_item_list[SI_AFS_MODE], mmd->afs_mode_string,

		status_item_list[SI_STREAM_START],
			(long unsigned)nmmd->stream_start.tv_sec,
			(long unsigned)nmmd->stream_start.tv_usec,
		status_item_list[SI_CURRENT_TIME],
			(long unsigned)now.tv_sec,
			(long unsigned)now.tv_usec,

		nmmd->afd.verbose_ls_output

	);
	free(flags);
	free(status);
	free(ut);
	return ret;
}

static int check_sender_args(int argc, char * const * argv, struct sender_command_data *scd)
{
	int i;
	/* this has to match sender.h */
	const char *subcmds[] = {"add", "delete", "allow", "deny", "on", "off", NULL};

	scd->sender_num = -1;
	if (argc < 2)
		return -E_COMMAND_SYNTAX;
	for (i = 0; senders[i].name; i++)
		if (!strcmp(senders[i].name, argv[1]))
			break;
	PARA_DEBUG_LOG("%d:%s\n", argc, argv[1]);
	if (!senders[i].name)
		return -E_COMMAND_SYNTAX;
	scd->sender_num = i;
	for (i = 0; subcmds[i]; i++)
		if (!strcmp(subcmds[i], argv[2]))
			break;
	if (!subcmds[i])
		return -E_COMMAND_SYNTAX;
	scd->cmd_num = i;
	if (!senders[scd->sender_num].client_cmds[scd->cmd_num])
		return -E_SENDER_CMD;
	switch (scd->cmd_num) {
	case SENDER_ON:
	case SENDER_OFF:
		if (argc != 3)
			return -E_COMMAND_SYNTAX;
		break;
	case SENDER_DENY:
	case SENDER_ALLOW:
		if (argc != 4 && argc != 5)
			return -E_COMMAND_SYNTAX;
		if (!inet_pton(AF_INET, argv[3], &scd->addr))
			return -E_COMMAND_SYNTAX;
		scd->netmask = 32;
		if (argc == 5) {
			scd->netmask = atoi(argv[4]);
			if (scd->netmask < 0 || scd->netmask > 32)
				return -E_COMMAND_SYNTAX;
		}
		break;
	case SENDER_ADD:
	case SENDER_DELETE:
		if (argc != 4 && argc != 5)
			return -E_COMMAND_SYNTAX;
		if (!inet_pton(AF_INET, argv[3], &scd->addr))
			return -E_COMMAND_SYNTAX;
		scd->port = -1;
		if (argc == 5) {
			scd->port = atoi(argv[4]);
			if (scd->port < 0 || scd->port > 65535)
				return -E_COMMAND_SYNTAX;
		}
		break;
	default:
		return -E_COMMAND_SYNTAX;
	}
	return 1;
}

int com_sender(int fd, int argc, char * const * argv)
{
	int i, ret;
	struct sender_command_data scd;

	if (argc < 2) {
		char *msg = NULL;
		for (i = 0; senders[i].name; i++) {
			char *tmp = make_message("%s%s\n",
				msg? msg : "", senders[i].name);
			free(msg);
			msg = tmp;
		}
		ret = send_buffer(fd, msg);
		free(msg);
		return ret;
	}
	ret = check_sender_args(argc, argv, &scd);
	if (ret < 0) {
		char *msg;
		if (scd.sender_num < 0)
			return ret;
		msg = senders[scd.sender_num].help();
		ret = send_buffer(fd, msg);
		free(msg);
		return ret;
	}
	for (i = 0; i < 10; i++) {
		mutex_lock(mmd_mutex);
		if (mmd->sender_cmd_data.cmd_num >= 0) {
			mutex_unlock(mmd_mutex);
			usleep(100 * 1000);
			continue;
		}
		mmd->sender_cmd_data = scd;
		mutex_unlock(mmd_mutex);
		break;
	}
	return (i < 10)? 1 : -E_LOCK;
}

/* server info */
int com_si(int fd, int argc, __a_unused char * const * argv)
{
	int i, ret;
	char *ut;
	char *sender_info = NULL, *sender_list = NULL;

	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	for (i = 0; senders[i].name; i++) {
		char *info = senders[i].info();
		sender_info = para_strcat(sender_info, info);
		free(info);
		sender_list = para_strcat(sender_list, senders[i].name);
		sender_list = para_strcat(sender_list, " ");
	}
	ut = uptime_str();
	ret = send_va_buffer(fd, "up: %s\nplayed: %u\n"
		"server_pid: %d\n"
		"afs_pid: %d\n"
		"connections (active/accepted/total): %u/%u/%u\n"
		"current loglevel: %i\n"
		"supported audio formats: %s\n"
		"supported senders: %s\n"
		"%s",
		ut, mmd->num_played,
		(int)getppid(),
		(int)mmd->afs_pid,
		mmd->active_connections,
		mmd->num_commands,
		mmd->num_connects,
		conf.loglevel_arg,
		supported_audio_formats(),
		sender_list,
		sender_info
	);
	mutex_unlock(mmd_mutex);
	free(ut);
	free(sender_list);
	free(sender_info);
	return ret;
}

/* version */
int com_version(int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	return send_buffer(fd, VERSION_TEXT("server")
		"built: " BUILD_DATE "\n"
		UNAME_RS ", " CC_VERSION "\n"
	);
}

/* stat */
int com_stat(int fd, int argc, char * const * argv)
{
	int ret, num = 0;/* status will be printed that many
			  * times. num <= 0 means: print forever
			  */
	struct misc_meta_data tmp, *nmmd = &tmp;
	char *s;

	signal(SIGUSR1, dummy);

	if (argc > 1)
		num = atoi(argv[1]);
	for (;;) {

		mmd_dup(nmmd);
		s = get_status(nmmd);
		ret = send_buffer(fd, s);
		free(s);
		if (ret < 0)
			goto out;
		ret = 1;
		if (num > 0 && !--num)
			goto out;
		sleep(50);
		if (getppid() == 1)
			return -E_SERVER_CRASH;
	}
out:
	return ret;
}

static int send_list_of_commands(int fd, struct server_command *cmd,
		const char *handler)
{
	int ret, i;

	for (i = 1; cmd->name; cmd++, i++) {
		char *perms = cmd_perms_itohuman(cmd->perms);
		ret = send_va_buffer(fd, "%s\t%s\t%s\t%s\n", cmd->name,
			handler,
			perms,
			cmd->description);
		free(perms);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/* returns string that must be freed by the caller */
static struct server_command *get_cmd_ptr(const char *name, char **handler)
{
	struct server_command *cmd;

	for (cmd = server_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("server"); /* server commands */
			return cmd;
		}
	/* not found, look for commands supported by afs */
	for (cmd = afs_cmds; cmd->name; cmd++)
		if (!strcmp(cmd->name, name)) {
			if (handler)
				*handler = para_strdup("afs");
			return cmd;
		}
	return NULL;
}

/* help */
int com_help(int fd, int argc, char * const * argv)
{
	struct server_command *cmd;
	char *perms, *handler;
	int ret;

	if (argc < 2) {
		/* no argument given, print list of commands */
		if ((ret = send_list_of_commands(fd, server_cmds, "server")) < 0)
			return ret;
		return send_list_of_commands(fd, afs_cmds, "afs");
	}
	/* argument given for help */
	cmd = get_cmd_ptr(argv[1], &handler);
	if (!cmd) {
		free(handler);
		return -E_BAD_CMD;
	}
	perms = cmd_perms_itohuman(cmd->perms);
	ret = send_va_buffer(fd,
		"%s - %s\n\n"
		"handler: %s\n"
		"permissions: %s\n"
		"usage: %s\n\n"
		"%s\n",
		argv[1],
		cmd->description,
		handler,
		perms,
		cmd->usage,
		cmd->help
	);
	free(perms);
	free(handler);
	return ret;
}

/* hup */
int com_hup(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGHUP);
	return 1;
}

/* term */
int com_term(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	kill(getppid(), SIGTERM);
	return 1;
}

int com_play(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;

}

/* stop */
int com_stop(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->new_vss_status_flags &= ~VSS_PLAYING;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* pause */
int com_pause(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (!vss_paused() && !vss_stopped()) {
		mmd->events++;
		mmd->new_vss_status_flags &= ~VSS_PLAYING;
		mmd->new_vss_status_flags &= ~VSS_NEXT;
	}
	mutex_unlock(mmd_mutex);
	return 1;
}

/* next */
int com_next(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	mmd->events++;
	mmd->new_vss_status_flags |= VSS_NEXT;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* nomore */
int com_nomore(__a_unused int fd, int argc, __a_unused char * const * argv)
{
	if (argc != 1)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	if (vss_playing() || vss_paused())
		mmd->new_vss_status_flags |= VSS_NOMORE;
	mutex_unlock(mmd_mutex);
	return 1;
}

/* ff */
int com_ff(__a_unused int fd, int argc, char * const * argv)
{
	long promille;
	int ret, backwards = 0;
	unsigned i;
	char c;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (!(ret = sscanf(argv[1], "%u%c", &i, &c)))
		return -E_COMMAND_SYNTAX;
	if (ret > 1 && c == '-')
		backwards = 1; /* jmp backwards */
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total || !mmd->afd.afhi.seconds_total)
		goto out;
	promille = (1000 * mmd->current_chunk) / mmd->afd.afhi.chunks_total;
	if (backwards)
		promille -= 1000 * i / mmd->afd.afhi.seconds_total;
	else
		promille += 1000 * i / mmd->afd.afhi.seconds_total;
	if (promille < 0)
		promille = 0;
	if (promille >  1000) {
		mmd->new_vss_status_flags |= VSS_NEXT;
		goto out;
	}
	mmd->repos_request = (mmd->afd.afhi.chunks_total * promille) / 1000;
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	mmd->events++;
	ret = 1;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

/* jmp */
int com_jmp(__a_unused int fd, int argc, char * const * argv)
{
	long unsigned int i;
	int ret;

	if (argc != 2)
		return -E_COMMAND_SYNTAX;
	if (sscanf(argv[1], "%lu", &i) <= 0)
		return -E_COMMAND_SYNTAX;
	mutex_lock(mmd_mutex);
	ret = -E_NO_AUDIO_FILE;
	if (!mmd->afd.afhi.chunks_total)
		goto out;
	if (i > 100)
		i = 100;
	PARA_INFO_LOG("jumping to %lu%%\n", i);
	mmd->repos_request = (mmd->afd.afhi.chunks_total * i + 50)/ 100;
	PARA_INFO_LOG("sent: %lu,  offset before jmp: %lu\n",
		mmd->chunks_sent, mmd->offset);
	mmd->new_vss_status_flags |= VSS_REPOS;
	mmd->new_vss_status_flags &= ~VSS_NEXT;
	ret = 1;
	mmd->events++;
out:
	mutex_unlock(mmd_mutex);
	return ret;
}

/*
 * check if perms are sufficient to exec a command having perms cmd_perms.
 * Returns 0 if perms are sufficient, -E_PERM otherwise.
 */
static int check_perms(unsigned int perms, struct server_command *cmd_ptr)
{
	PARA_DEBUG_LOG("checking permissions\n");
	return (cmd_ptr->perms & perms) < cmd_ptr->perms ? -E_PERM : 0;
}

/*
 * Parse first string from *cmd and lookup in table of valid commands.
 * On error, NULL is returned.
 */
static struct server_command *parse_cmd(const char *cmdstr)
{
	char buf[255];
	int n = 0;

	sscanf(cmdstr, "%200s%n", buf, &n);
	if (!n)
		return NULL;
	buf[n] = '\0';
	return get_cmd_ptr(buf, NULL);
}

static void init_rc4_keys(void)
{
	int i;

	for (i = 0; i < 2 * RC4_KEY_LEN; i++)
		rc4_buf[i] = para_random(256);
	PARA_DEBUG_LOG("rc4 keys initialized (%u:%u)\n",
		(unsigned char) rc4_buf[0],
		(unsigned char) rc4_buf[RC4_KEY_LEN]);
	RC4_set_key(&rc4_recv_key, RC4_KEY_LEN, rc4_buf);
	RC4_set_key(&rc4_send_key, RC4_KEY_LEN, rc4_buf + RC4_KEY_LEN);
}

static void rc4_recv(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, __a_unused void *private_data)
{
	RC4(&rc4_recv_key, len, indata, outdata);
}

static void rc4_send(unsigned long len, const unsigned char *indata,
		unsigned char *outdata, __a_unused void *private_data)
{
	RC4(&rc4_send_key, len, indata, outdata);
}

static int read_command(int fd, char **result)
{
	int ret;
	char buf[4096];
	char *command = NULL;

	for (;;) {
		size_t numbytes;
		char *p;

		ret = recv_buffer(fd, buf, sizeof(buf));
		if (ret < 0)
			goto out;
		if (!ret)
			break;
		numbytes = ret;
		ret = -E_COMMAND_SYNTAX;
		if (command && numbytes + strlen(command) > MAX_COMMAND_LEN) /* DOS */
			goto out;
		command = para_strcat(command, buf);
		p = strstr(command, EOC_MSG);
		if (p) {
			*p = '\0';
			break;
		}
	}
	ret = command? 1 : -E_COMMAND_SYNTAX;
out:
	if (ret < 0)
		free(command);
	else
		*result = command;
	return ret;

}

/**
 * Perform user authentication and execute a command.
 *
 * \param fd The file descriptor to send output to.
 * \param peername Identifies the connecting peer.
 *
 * Whenever para_server accepts an incoming tcp connection on
 * the port it listens on, it forks and the resulting child
 * calls this function.
 *
 * An RSA-based challenge/response is used to authenticate
 * the peer. It that authentication succeeds, a random RC4
 * session key is generated and sent back to the peer,
 * encrypted with its RSA public key.  From this point on,
 * all transfers are crypted with this session key.
 *
 * Next it is checked if the peer supplied  a valid server command or a command
 * for the audio file selector.  If yes, and if the user has sufficient
 * permissions to execute that command, the function calls the corresponding
 * command handler which does argument checking and further processing.
 *
 * In order to cope with a DOS attacks, a timeout is set up
 * which terminates the function if the connection was not
 * authenticated when the timeout expires.
 *
 * \sa alarm(2), rc4(3), crypt.c, crypt.h
 */
__noreturn void handle_connect(int fd, const char *peername)
{
	int ret, argc, use_rc4 = 0;
	char buf[4096];
	unsigned char crypt_buf[MAXLINE];
	struct user *u;
	struct server_command *cmd = NULL;
	long unsigned challenge_nr, chall_response;
	char **argv = NULL;
	char *p, *command = NULL;
	size_t numbytes;

	signal(SIGCHLD, SIG_IGN);
	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);

	/* we need a blocking fd here as recv() might return EAGAIN otherwise. */
	ret = mark_fd_blocking(fd);
	if (ret < 0)
		goto err_out;
	challenge_nr = random();
	/* send Welcome message */
	ret = send_va_buffer(fd, "This is para_server, version "
		PACKAGE_VERSION  ".\n" );
	if (ret < 0)
		goto err_out;
	/* recv auth request line */
	ret = recv_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto err_out;
	if (ret <= 6) {
		ret = -E_AUTH;
		goto err_out;
	}
	numbytes = ret;
	ret = -E_AUTH;
	if (strncmp(buf, "auth ", 5))
		goto err_out;

	if (numbytes < 9 || strncmp(buf, "auth rc4 ", 9))
		p = buf + 5; /* client version < 0.2.6 */
	else {
		p = buf + 9; /* client version >= 0.2.6 */
		use_rc4 = 1;
	}
	PARA_DEBUG_LOG("received %s request for user %s\n",
		use_rc4? "rc4" : "auth", p);
	ret = -E_BAD_USER;
	u = lookup_user(p);
	if (!u)
		goto err_out;
	ret = para_encrypt_challenge(u->rsa, challenge_nr, crypt_buf);
	if (ret <= 0)
		goto err_out;
	numbytes = ret;
	PARA_DEBUG_LOG("sending %zu byte challenge\n", numbytes);
	/* We can't use send_buffer here since buf may contain null bytes */
	ret = send_bin_buffer(fd,(char *) crypt_buf, numbytes);
	if (ret < 0)
		goto net_err;
	/* recv decrypted number */
	ret = recv_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto net_err;
	numbytes = ret;
	ret = -E_AUTH;
	if (!numbytes)
		goto net_err;
	if (sscanf(buf, CHALLENGE_RESPONSE_MSG "%lu", &chall_response) < 1
			|| chall_response != challenge_nr)
		goto err_out;
	/* auth successful, send 'Proceed' message */
	PARA_INFO_LOG("good auth for %s (%lu)\n", u->name, challenge_nr);
	sprintf(buf, "%s", PROCEED_MSG);
	if (use_rc4) {
		init_rc4_keys();
		ret = para_encrypt_buffer(u->rsa, rc4_buf, 2 * RC4_KEY_LEN,
			(unsigned char *)buf + PROCEED_MSG_LEN + 1);
		if (ret <= 0)
			goto err_out;
		numbytes = ret + strlen(PROCEED_MSG) + 1;
	} else
		numbytes = strlen(buf);
	ret = send_bin_buffer(fd, buf, numbytes);
	if (ret < 0)
		goto net_err;
	if (use_rc4)
		enable_crypt(fd, rc4_recv, rc4_send, NULL);
	ret = read_command(fd, &command);
	if (ret == -E_COMMAND_SYNTAX)
		goto err_out;
	if (ret < 0)
		goto net_err;
	ret = -E_BAD_CMD;
	cmd = parse_cmd(command);
	if (!cmd)
		goto err_out;
	/* valid command, check permissions */
	ret = check_perms(u->perms, cmd);
	if (ret < 0)
		goto err_out;
	/* valid command and sufficient perms */
	alarm(0);
	argc = split_args(command, &argv, "\n");
	mutex_lock(mmd_mutex);
	mmd->num_commands++;
	mutex_unlock(mmd_mutex);
	PARA_NOTICE_LOG("calling com_%s() for %s@%s\n", cmd->name, u->name,
			peername);
	ret = cmd->handler(fd, argc, argv);
	if (ret >= 0)
		goto out;
err_out:
	send_va_buffer(fd, "%s\n", para_strerror(-ret));
net_err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
out:
	free(command);
	free(argv);
	mutex_lock(mmd_mutex);
	if (cmd && (cmd->perms & AFS_WRITE) && ret >= 0)
		mmd->events++;
	mmd->active_connections--;
	mutex_unlock(mmd_mutex);
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
