/*
 * Copyright (C) 2005-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file http_recv.c paraslash's http receiver */

#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "http.h"
#include "list.h"
#include "sched.h"
#include "recv.h"
#include "http_recv.cmdline.h"
#include "net.h"
#include "string.h"
#include "fd.h"

/** the output buffer size of the http receiver */
#define BUFSIZE (32 * 1024)

/**
 * the possible states of a http receiver node
 *
 * \sa receiver_node
 */
enum http_recv_status {HTTP_CONNECTED, HTTP_SENT_GET_REQUEST, HTTP_STREAMING};

/**
 * data specific to the http receiver
 *
 * Each running instance of the http receiver reserves space for one such struct.
 */
struct private_http_recv_data {
/**
 *
 *
 * the current status of the http receiver node
 *
 * It gets initialized to \p HTTP_CONNECTED by the open function of the
 * http receiver.
 *
 * \sa receiver::open, receiver_node.
 */
	enum http_recv_status status;
/**
 *
 *
 * the file descriptor used for receiving the http stream
 *
 * The pre_select function of the http receiver adds this file descriptor to
 * the set of file decriptors which are checked for reading/writing (depending
 * on the current status) by the select loop of the application (para_audiod or
 * para_recv).
 *
 * The post_select function of the http receiver uses \a fd, if ready, to
 * establish the http connection, and updates \a status according to the new
 * state of the connection.  As soon as \a status is \p HTTP_STREAMING, \a fd is
 * going to be only checked for reading. If data is available, it is read into
 * the output buffer of the receiver node by post_select.
 *
 * \sa receiver::pre_select receiver::post_select receiver_node, http_recv_status
 */
	int fd;
};

static void http_shutdown(void)
{
	return;
}

static char *make_request_msg(void)
{
	char *ret, *hn = para_hostname();
	ret = make_message("%s1.0\nHost: %s\nUser-Agent: para_recv/%s\n\n\n",
		HTTP_GET_MSG, hn, PACKAGE_VERSION);
	free(hn);
	return ret;
}

static void http_recv_pre_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_http_recv_data *phd = rn->private_data;

	t->error = 0;
	if  (phd->status == HTTP_CONNECTED)
		para_fd_set(phd->fd, &s->wfds, &s->max_fileno);
	else
		para_fd_set(phd->fd, &s->rfds, &s->max_fileno);
}


static void http_recv_post_select(struct sched *s, struct task *t)
{
	struct receiver_node *rn = container_of(t, struct receiver_node, task);
	struct private_http_recv_data *phd = rn->private_data;

	if (rn->output_error && *rn->output_error < 0) {
		t->error = *rn->output_error;
		return;
	}
	if  (phd->status == HTTP_CONNECTED) {
		char *rq;
		if (!FD_ISSET(phd->fd, &s->wfds))
			return;
		rq = make_request_msg();
		PARA_INFO_LOG("sending http request\n");
		t->error = send_va_buffer(phd->fd, "%s", rq);
		free(rq);
		if (t->error >= 0)
			phd->status = HTTP_SENT_GET_REQUEST;
		return;
	}
	if (!FD_ISSET(phd->fd, &s->rfds))
		return;
	if (phd->status == HTTP_SENT_GET_REQUEST) {
		t->error = recv_pattern(phd->fd, HTTP_OK_MSG, strlen(HTTP_OK_MSG));
		if (t->error >= 0) {
			PARA_INFO_LOG("received ok msg, streaming\n");
			phd->status = HTTP_STREAMING;
		}
		return;
	}
	if (rn->loaded >= BUFSIZE) {
		t->error = -E_HTTP_RECV_OVERRUN;
		return;
	}
	t->error = recv_bin_buffer(phd->fd, rn->buf + rn->loaded,
		BUFSIZE - rn->loaded);
	if (t->error > 0) {
		rn->loaded += t->error;
		return;
	}
	if (!t->error)
		t->error = -E_RECV_EOF;
}

static void http_recv_close(struct receiver_node *rn)
{
	struct private_http_recv_data *phd = rn->private_data;
	close(phd->fd);
	free(rn->buf);
	free(rn->private_data);
}

static void *http_recv_parse_config(int argc, char **argv)
{
	struct http_recv_args_info *tmp = para_calloc(sizeof(struct http_recv_args_info));

	if (!http_recv_cmdline_parser(argc, argv, tmp))
		return tmp;
	free(tmp);
	return NULL;
}

static int http_recv_open(struct receiver_node *rn)
{
	struct private_http_recv_data *phd;
	struct http_recv_args_info *conf = rn->conf;
	int fd, ret = makesock(AF_UNSPEC, IPPROTO_TCP, 0, conf->host_arg,
		conf->port_arg);

	if (ret < 0)
		return ret;
	fd = ret;
	ret = mark_fd_nonblocking(fd);
	if (ret < 0) {
		close(fd);
		return ret;
	}
	rn->buf = para_calloc(BUFSIZE);
	rn->private_data = phd = para_calloc(sizeof(struct private_http_recv_data));
	phd->fd = fd;
	phd->status = HTTP_CONNECTED;
	return 1;
}

/**
 * the init function of the http receiver
 *
 * \param r pointer to the receiver struct to initialize
 *
 * Just initialize all function pointers of \a r.
 */
void http_recv_init(struct receiver *r)
{
	r->open = http_recv_open;
	r->close = http_recv_close;
	r->pre_select = http_recv_pre_select;
	r->post_select = http_recv_post_select;
	r->shutdown = http_shutdown;
	r->parse_config = http_recv_parse_config;
}
