/*
 * Copyright (C) 2003-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file mp3_afh.c para_server's mp3 audio format handler */

/*
 * This file is based in part on mp3tech.c and mp3tech.h, Copyright (C)
 * 2000-2001 Cedric Tefft <cedric@earthling.net>, which in turn is based
 * in part on
 *
 *	* MP3Info 0.5 by Ricardo Cerqueira <rmc@rccn.net>
 *	* MP3Stat 0.9 by Ed Sweetman <safemode@voicenet.com> and
 *			 Johannes Overmann <overmann@iname.com>
 */

#include "para.h"
#include "error.h"
#include "afh.h"
#include "string.h"
#include "afs.h"
#include "server.h"

/** \cond some defines and structs which are only used in this file */

/*
 * MIN_CONSEC_GOOD_FRAMES defines how many consecutive valid MP3 frames we need
 * to see before we decide we are looking at a real MP3 file
 */
#define MIN_CONSEC_GOOD_FRAMES 4
#define FRAME_HEADER_SIZE 4
#define MIN_FRAME_SIZE 21

struct mp3header {
	unsigned long sync;
	unsigned int version;
	unsigned int layer;
	unsigned int crc;
	unsigned int bitrate;
	unsigned int freq;
	unsigned int padding;
	unsigned int mode;
	unsigned int copyright;
	unsigned int original;
	unsigned int emphasis;
};

/** \endcond */
static const int frequencies[3][4] = {
	{22050,24000,16000,50000}, /* MPEG 2.0 */
	{44100,48000,32000,50000}, /* MPEG 1.0 */
	{11025,12000,8000,50000} /* MPEG 2.5 */
};

static const int mp3info_bitrate[2][3][14] = {
{ /* MPEG 2.0 */
	{32,48,56,64,80,96,112,128,144,160,176,192,224,256}, /* layer 1 */
	{8,16,24,32,40,48,56,64,80,96,112,128,144,160}, /* layer 2 */
	{8,16,24,32,40,48,56,64,80,96,112,128,144,160} /* layer 3 */
},

{ /* MPEG 1.0 */
	{32,64,96,128,160,192,224,256,288,320,352,384,416,448}, /* layer 1 */
	{32,48,56,64,80,96,112,128,160,192,224,256,320,384}, /* layer 2 */
	{32,40,48,56,64,80,96,112,128,160,192,224,256,320} /* layer 3 */
}
};

static const int frame_size_index[] = {24000, 72000, 72000};
static const char *mode_text[] = {"stereo", "joint stereo", "dual channel", "mono", "invalid"};

#ifdef HAVE_LIBID3TAG

#include <id3tag.h>

static char *get_latin1(id3_ucs4_t const *string)
{
	if (!string)
		return NULL;
	return (char *)id3_ucs4_latin1duplicate(string);
}

char *get_stringlist(union id3_field *field)
{
	unsigned int k, nstrings = id3_field_getnstrings(field);
	char *result = NULL;

	for (k = 0; k < nstrings; k++) {
		char *tmp = (char *)get_latin1(id3_field_getstrings(field, k));
		if (result) {
			char *tmp2 = result;
			result = make_message("%s %s", tmp2, tmp);
			free(tmp);
			free(tmp2);
		} else
			result = tmp;
	}
	return result;
}

static char *get_string(union id3_field *field)
{
	id3_ucs4_t const *string = id3_field_getfullstring(field);

	return get_latin1(string);
}

#define FOR_EACH_FIELD(f, j, fr) for (j = 0; j < (fr)->nfields && \
	(f = id3_frame_field((fr), j)); j++)

static char *get_strings(struct id3_frame *fr)
{
	int j;
	union id3_field *field;

	FOR_EACH_FIELD(field, j, fr) {
		enum id3_field_type type = id3_field_type(field);

		if (type == ID3_FIELD_TYPE_STRINGLIST)
			return get_stringlist(field);
		if (type == ID3_FIELD_TYPE_STRINGFULL)
			return get_string(field);
	}
	return NULL;
}

static char *mp3_get_id3(__a_unused unsigned char *map,
		__a_unused size_t numbytes, int fd)
{
	int i;
	struct id3_tag *id3_t;
	char *title = NULL, *artist = NULL, *album = NULL, *year = NULL,
		*comment = NULL, *result;
	struct id3_file *id3_f = id3_file_fdopen(fd, ID3_FILE_MODE_READONLY);

	if (!id3_f)
		goto no_tag;
	id3_t = id3_file_tag(id3_f);
	if (!id3_t)
		goto no_tag;
	for (i = 0; i < id3_t->nframes; i++) {
		struct id3_frame *fr = id3_t->frames[i];
		if (!strcmp(fr->id, "TIT2")) {
			if (!title)
				title = get_strings(fr);
			continue;
		}
		if (!strcmp(fr->id, "TPE1")) {
			if (!artist)
				artist = get_strings(fr);
			continue;
		}
		if (!strcmp(fr->id, "TALB")) {
			if (!album)
				album = get_strings(fr);
			continue;
		}
		if (!strcmp(fr->id, "TDRC")) {
			if (!year)
				year = get_strings(fr);
			continue;
		}
		if (!strcmp(fr->id, "COMM")) {
			if (!comment)
				comment = get_strings(fr);
			continue;
		}
	}
	id3_file_close(id3_f);
	result = make_taginfo(title, artist, album, year, comment);
	free(title);
	free(artist);
	free(album);
	free(year);
	free(comment);
	return result;
no_tag:
	if (id3_f)
		id3_file_close(id3_f);
	return make_message("%s: (no id3 v1/v2 tag)\n%s:\n",
		status_item_list[SI_TAGINFO1],
		status_item_list[SI_TAGINFO2]);
}

#else /* HAVE_LIBID3TAG */

/*
 * Remove trailing whitespace from the end of a string
 */
static char *unpad(char *string)
{
	char *pos = string + strlen(string) - 1;
	while (para_isspace(pos[0]))
		(pos--)[0] = 0;
	return string;
}

static char *mp3_get_id3(unsigned char *map, size_t numbytes, __a_unused int fd)
{
	char title[31], artist[31], album[31], year[5], comment[31];
	off_t fpos;

	if (numbytes < 128 || strncmp("TAG", (char *)map + numbytes - 128, 3)) {
		PARA_DEBUG_LOG("no id3 v1 tag\n");
		return make_message("%s: (no id3 v1 tag)\n%s:\n",
			status_item_list[SI_TAGINFO1],
			status_item_list[SI_TAGINFO2]);
	}
	fpos = numbytes - 125;
	memcpy(title, map + fpos, 30);
	fpos += 30;
	title[30] = '\0';
	memcpy(artist, map + fpos, 30);
	fpos += 30;
	artist[30] = '\0';
	memcpy(album, map + fpos, 30);
	fpos += 30;
	album[30] = '\0';
	memcpy(year, map + fpos, 4);
	fpos += 4;
	year[4] = '\0';
	memcpy(comment, map + fpos, 30);
	comment[30] = '\0';
	unpad(title);
	unpad(artist);
	unpad(album);
	unpad(year);
	unpad(comment);
	return make_taginfo(title, artist, album, year, comment);
}
#endif /* HAVE_LIBID3TAG */

static int header_frequency(struct mp3header *h)
{
	if (h->version > 2 || h->freq > 3)
		return -E_HEADER_FREQ;
	return frequencies[h->version][h->freq];
}

static const char *header_mode(struct mp3header *h)
{
	if (h->mode > 4)
		h->mode = 4; /* invalid */
	return mode_text[h->mode];
}

static int header_channels(struct mp3header *h)
{
	if (h->mode > 3)
		return 0;
	if (h->mode < 3)
		return 2;
	return 1;
}

static int header_bitrate(struct mp3header *h)
{
	if (!h->layer || h->layer > 3 || h->bitrate > 14 || !h->bitrate)
		return -E_HEADER_BITRATE;
	return mp3info_bitrate[h->version & 1][3 - h->layer][h->bitrate - 1];
}

static int frame_length(struct mp3header *header)
{
	int hb, hf = header_frequency(header);

	if (hf < 0)
		return hf;
	hb = header_bitrate(header);
	if (hb < 0)
		return hb;
	if (header->sync != 0xFFE || header->layer > 3)
		return -E_FRAME;
	return frame_size_index[3 - header->layer] *
		((header->version & 1) + 1) * hb / hf
		+ header->padding;
}

static int compare_headers(struct mp3header *h1,struct mp3header *h2)
{
	if ((*(unsigned int*)h1) == (*(unsigned int*)h2))
		return 1;
	if ((h1->version == h2->version) &&
			(h1->layer == h2->layer) &&
			(h1->crc == h2->crc) &&
			(h1->freq == h2->freq) &&
			(h1->mode == h2->mode) &&
			(h1->copyright == h2->copyright) &&
			(h1->original == h2->original) &&
			(h1->emphasis == h2->emphasis))
		return 1;
	return 0;
}

/*
 * get next MP3 frame header.
 *
 * On success, the header frame length is returned and the given header
 * structure that is filled in.  A return value of zero means that we did not
 * retrieve a valid frame header, and a negative return value indicates an
 * error.
 */
static int get_header(unsigned char *map, size_t numbytes, off_t *fpos,
	struct mp3header *header)
{
	int fl, ret;

	if (*fpos + FRAME_HEADER_SIZE > numbytes) {
		*fpos = numbytes - 1;
		header->sync = 0;
		return 0;
	}
	header->layer = (map[*fpos + 1] >> 1) & 3;
	header->sync = (((int)map[*fpos]<<4) | ((int)(map[*fpos + 1]&0xE0)>>4));
	if (map[*fpos + 1] & 0x10)
		header->version = (map[*fpos + 1] >> 3) & 1;
	else
		header->version = 2;
	if ((header->sync != 0xFFE) || (header->layer != 1)) {
		ret = 0;
		header->sync = 0;
		goto out;
	}
	header->crc = map[*fpos + 1] & 1;
	header->bitrate = (map[*fpos + 2] >> 4) & 0x0F;
	header->freq = (map[*fpos + 2] >> 2) & 0x3;
	header->padding = (map[*fpos + 2] >>1) & 0x1;
	header->mode = (map[*fpos + 3] >> 6) & 0x3;
	fl = frame_length(header);
	ret = (fl >= MIN_FRAME_SIZE)? fl : -E_FRAME_LENGTH;
out:
	*fpos += FRAME_HEADER_SIZE;
	return ret;
}

/*
 * find the next mp3 header
 *
 * Return the length of the next frame header or zero if the end of the file is
 * reached.
 */
static int mp3_seek_next_header(unsigned char *map, size_t numbytes, off_t *fpos,
	struct mp3header *result)
{
	int k, l = 0, first_len;
	struct mp3header h, h2;
	long valid_start = 0;

	for (; *fpos < numbytes; (*fpos)++) {
		if (map[*fpos] != 0xff)
			continue;
		valid_start = *fpos;
		first_len = get_header(map, numbytes, fpos, &h);
		if (first_len <= 0)
			continue;
		*fpos += first_len - FRAME_HEADER_SIZE;
		for (k = 1; k < MIN_CONSEC_GOOD_FRAMES; k++) {
			if ((l = get_header(map, numbytes, fpos, &h2)) <= 0)
				break;
			if (!compare_headers(&h, &h2))
				break;
			*fpos += l - FRAME_HEADER_SIZE;
		}
		if (k == MIN_CONSEC_GOOD_FRAMES) {
			*fpos = valid_start;
			*result = h2;
			return first_len;
		}
	}
	return 0;
}

static int find_valid_start(unsigned char *map, size_t numbytes, off_t *fpos,
	struct mp3header *header)
{
	int frame_len;

	frame_len = get_header(map, numbytes, fpos, header);
	if (frame_len < 0)
		return frame_len;
	if (!frame_len) {
		frame_len = mp3_seek_next_header(map, numbytes, fpos, header);
		if (frame_len <= 0)
			return frame_len;
	} else
		*fpos -= FRAME_HEADER_SIZE;
	if (frame_len <= 1)
		return -E_FRAME_LENGTH;
	return frame_len;
}

static int mp3_read_info(unsigned char *map, size_t numbytes, int fd,
		struct afh_info *afhi)
{
	long freq_avg = 0, br_avg = 0;
	int ret, len = 0, old_br = -1, vbr = 0;
	struct timeval total_time = {0, 0};
	unsigned chunk_table_size = 1000; /* gets increased on demand */
	off_t fpos = 0;
	struct mp3header header;
	char *taginfo;

	afhi->chunks_total = 0;
	afhi->chunk_table = para_malloc(chunk_table_size * sizeof(size_t));
	taginfo = mp3_get_id3(map, numbytes, fd);
	while (1) {
		unsigned long freq, br, fl;
		struct timeval tmp, cct; /* current chunk time */
		fpos += len;
		len = find_valid_start(map, numbytes, &fpos, &header);
		if (len <= 0)
			break;
		ret = header_frequency(&header);
		if (ret < 0)
			continue;
		freq = ret;
		ret = header_bitrate(&header);
		if (ret < 0)
			continue;
		br = ret;
		ret = frame_length(&header);
		if (ret < 0)
			continue;
		fl = ret;
		tmp.tv_sec = fl;
		tmp.tv_usec = 0;
		tv_divide(br * 125, &tmp, &cct);
		tv_add(&cct, &total_time, &tmp);
		total_time = tmp;
		//PARA_DEBUG_LOG("%s: br: %d, freq: %d, fl: %d, cct: %lu\n", __func__, br, freq, fl, cct.tv_usec);
		if (afhi->chunks_total >= chunk_table_size) {
			chunk_table_size *= 2;
			afhi->chunk_table = para_realloc(afhi->chunk_table,
				chunk_table_size * sizeof(size_t));
		}
		afhi->chunk_table[afhi->chunks_total] = fpos;
//		if (afhi->chunks_total < 10 || !(afhi->chunks_total % 1000))
//			PARA_INFO_LOG("chunk #%lu: %zd\n", afhi->chunks_total,
//				afhi->chunk_table[afhi->chunks_total]);
		afhi->chunks_total++;
		if (afhi->chunks_total == 1) {
			freq_avg = freq;
			br_avg = br;
			old_br = br;
			continue;
		}
		freq_avg += ((long)freq - freq_avg) / ((long)afhi->chunks_total + 1);
		br_avg += ((long)br - br_avg) / ((long)afhi->chunks_total + 1);
		if (old_br != br)
			vbr = 1;
		old_br = br;
	}
	ret = -E_MP3_INFO;
	if (!afhi->chunks_total || !freq_avg || !br_avg)
		goto err_out;
	afhi->chunk_table[afhi->chunks_total] = numbytes - 1;
	afhi->bitrate = br_avg;
	afhi->frequency = freq_avg;
	afhi->channels = header_channels(&header);
	afhi->seconds_total = (tv2ms(&total_time) + 500) / 1000;
	tv_divide(afhi->chunks_total, &total_time, &afhi->chunk_tv);
	PARA_DEBUG_LOG("%lu chunks, each %lums\n", afhi->chunks_total,
		tv2ms(&afhi->chunk_tv));
	tv_scale(3, &afhi->chunk_tv, &afhi->eof_tv);
	PARA_DEBUG_LOG("eof timeout: %lu\n", tv2ms(&afhi->eof_tv));
	afhi->info_string = make_message("%s: %cbr, %s\n%s",
		status_item_list[SI_AUDIO_FILE_INFO], vbr? 'v' : 'c',
		header_mode(&header), taginfo);
	free(taginfo);
	return 1;
err_out:
	free(taginfo);
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	free(afhi->chunk_table);
	return ret;
}

/*
 * Read mp3 information from audio file
 */
static int mp3_get_file_info(char *map, size_t numbytes, int fd,
		struct afh_info *afhi)
{
	int ret;

	ret = mp3_read_info((unsigned char *)map, numbytes, fd, afhi);
	if (ret < 0)
		return ret;
	if (afhi->seconds_total < 2 || !afhi->chunks_total)
		return -E_MP3_INFO;
	return 1;
}

static const char* mp3_suffixes[] = {"mp3", NULL};

/**
 * the init function of the mp3 audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void mp3_init(struct audio_format_handler *afh)
{
	afh->get_file_info = mp3_get_file_info;
	afh->suffixes = mp3_suffixes;
}
