/*
 * Copyright (C) 1997-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file vss.c The virtual streaming system.
 *
 * This contains the audio streaming code of para_server which is independent
 * of the current audio format, audio file selector and of the activated
 * senders.
 */

#include <dirent.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "net.h"
#include "server.cmdline.h"
#include "vss.h"
#include "list.h"
#include "send.h"
#include "ipc.h"
#include "fd.h"
#include "sched.h"

extern struct misc_meta_data *mmd;

extern void dccp_send_init(struct sender *);
extern void http_send_init(struct sender *);
extern void ortp_send_init(struct sender *);

/** The list of supported senders. */
struct sender senders[] = {
	{
		.name = "http",
		.init = http_send_init,
	},
	{
		.name = "dccp",
		.init = dccp_send_init,
	},
#ifdef HAVE_ORTP
	{
		.name = "ortp",
		.init = ortp_send_init,
	},
#endif
	{
		.name = NULL,
	}
};

/** The possible states of the afs socket. */
enum afs_socket_status {
	/** Socket is inactive. */
	AFS_SOCKET_READY,
	/** Socket fd was included in the write fd set for select(). */
	AFS_SOCKET_CHECK_FOR_WRITE,
	/** vss wrote a request to the socket and waits for reply from afs. */
	AFS_SOCKET_AFD_PENDING
};

/** The task structure for the virtual streaming system. */
struct vss_task {
	struct timeval announce_tv;
	struct timeval data_send_barrier;
	struct timeval eof_barrier;
	struct timeval autoplay_barrier;
	int afs_socket;
	enum afs_socket_status afsss;
	char *map;
	struct task task;
	const char *header_buf;
	size_t header_len;
};

/**
 * Check if vss status flag \a P (playing) is set.
 *
 * \return Greater than zero if playing, zero otherwise.
 *
 */
unsigned int vss_playing(void)
{
	return mmd->new_vss_status_flags & VSS_PLAYING;
}

/**
 * Check if the \a N (next) status flag is set.
 *
 * \return Greater than zero if set, zero if not.
 *
 */
unsigned int vss_next(void)
{
	return mmd->new_vss_status_flags & VSS_NEXT;
}

/**
 * Check if a reposition request is pending.
 *
 * \return Greater than zero if true, zero otherwise.
 *
 */
unsigned int vss_repos(void)
{
	return mmd->new_vss_status_flags & VSS_REPOS;
}

/**
 * Check if the vss is currently paused.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_paused(void)
{
	return !(mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

/**
 * Check if the vss is currently stopped.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_stopped(void)
{
	return (mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

static int chk_barrier(const char *bname, const struct timeval *barrier,
		struct timeval *diff, int print_log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (print_log && ms)
		PARA_DEBUG_LOG("%s barrier: %lims left\n", bname, ms);
	return -1;
}

/*
 * != NULL: timeout for next chunk
 * NULL: nothing to do
 */
static struct timeval *vss_compute_timeout(struct vss_task *vsst)
{
	static struct timeval the_timeout;
	struct timeval next_chunk;

	if (vss_next() && vsst->map) {
		/* only sleep a bit, nec*/
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 100;
		return &the_timeout;
	}
	if (chk_barrier("autoplay_delay", &vsst->autoplay_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("eof", &vsst->eof_barrier, &the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("data send", &vsst->data_send_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (!vss_playing() || !vsst->map)
		return NULL;
	compute_chunk_time(mmd->chunks_sent, &mmd->afd.afhi.chunk_tv,
		&mmd->stream_start, &next_chunk);
	if (chk_barrier("chunk", &next_chunk, &the_timeout, 0) < 0)
		return &the_timeout;
	/* chunk is due or bof */
	the_timeout.tv_sec = 0;
	the_timeout.tv_usec = 0;
	return &the_timeout;
}

static void vss_eof(struct vss_task *vsst)
{
	mmd->stream_start = *now;
	if (!vsst->map)
		return;
	if (mmd->new_vss_status_flags & VSS_NOMORE)
		mmd->new_vss_status_flags = VSS_NEXT;
	tv_add(&mmd->afd.afhi.eof_tv, now, &vsst->eof_barrier);
	para_munmap(vsst->map, mmd->size);
	vsst->map = NULL;
	mmd->chunks_sent = 0;
	mmd->offset = 0;
	mmd->afd.afhi.seconds_total = 0;
	mmd->afd.afhi.chunk_tv.tv_sec = 0;
	mmd->afd.afhi.chunk_tv.tv_usec = 0;
	free(mmd->afd.afhi.chunk_table);
	mmd->afd.afhi.chunk_table = NULL;
	free(mmd->afd.afhi.info_string);
	mmd->afd.afhi.info_string = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_AUDIO_FILE_INFO],
		status_item_list[SI_TAGINFO1], status_item_list[SI_TAGINFO2]);
	make_empty_status_items(mmd->afd.verbose_ls_output);
	mmd->mtime = 0;
	mmd->size = 0;
	mmd->events++;
}

/**
 * Get the list of all supported audio formats.
 *
 * \return Aa space separated list of all supported audio formats
 * It is not allocated at runtime, i.e. there is no need to free
 * the returned string in the caller.
 */
const char *supported_audio_formats(void)
{
	return SUPPORTED_AUDIO_FORMATS;
}

/**
 * Get the chunk time of the current audio file.
 *
 * \return A pointer to a struct containing the chunk time, or NULL,
 * if currently no audio file is selected.
 */
struct timeval *vss_chunk_time(void)
{
	if (mmd->afd.afhi.chunk_tv.tv_sec == 0 &&
			mmd->afd.afhi.chunk_tv.tv_usec == 0)
		return NULL;
	return &mmd->afd.afhi.chunk_tv;
}

static int need_to_request_new_audio_file(struct vss_task *vsst)
{
	if (vsst->map) /* have audio file */
		return 0;
	if (!vss_playing()) /* don't need one */
		return 0;
	if (mmd->new_vss_status_flags & VSS_NOMORE)
		return 0;
	if (vsst->afsss == AFS_SOCKET_AFD_PENDING) /* already requested one */
		return 0;
	return 1;
}

/**
 * Compute the timeout for para_server's main select-loop.
 *
 * This function gets called from para_server to determine the timeout value
 * for its main select loop.
 *
 * \param s Pointer to the server scheduler.
 * \param t Pointer to the vss task structure.
 *
 * Before the timeout is computed, the current vss status flags are evaluated
 * and acted upon by calling appropriate functions from the lower layers.
 * Possible actions include
 *
 * 	- request a new audio file from afs,
 *	- shutdown of all senders (stop/pause command),
 *	- reposition the stream (ff/jmp command).
 */
static void vss_pre_select(struct sched *s, struct task *t)
{
	int i;
	struct timeval *tv, diff;
	struct vss_task *vsst = container_of(t, struct vss_task, task);

	if (!vsst->map || vss_next() || vss_paused() || vss_repos())
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
	if (vss_next())
		vss_eof(vsst);
	else if (vss_paused()) {
		if (mmd->chunks_sent)
			tv_add(&mmd->afd.afhi.eof_tv, now, &vsst->eof_barrier);
		mmd->chunks_sent = 0;
	} else if (vss_repos()) {
		tv_add(now, &vsst->announce_tv, &vsst->data_send_barrier);
		tv_add(&mmd->afd.afhi.eof_tv, now, &vsst->eof_barrier);
		mmd->chunks_sent = 0;
		mmd->current_chunk = mmd->repos_request;
		mmd->new_vss_status_flags &= ~VSS_REPOS;
	}

	if (need_to_request_new_audio_file(vsst)) {
		PARA_DEBUG_LOG("ready and playing, but no audio file\n");
		para_fd_set(vsst->afs_socket, &s->wfds, &s->max_fileno);
		vsst->afsss = AFS_SOCKET_CHECK_FOR_WRITE;
	} else
		para_fd_set(vsst->afs_socket, &s->rfds, &s->max_fileno);
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].pre_select)
			continue;
		senders[i].pre_select(&s->max_fileno, &s->rfds, &s->wfds);
	}
	tv = vss_compute_timeout(vsst);
	if (tv && tv_diff(tv, &s->timeout, &diff) < 0)
		s->timeout = *tv;
}

static int recv_afs_msg(int afs_socket, int *fd, uint32_t *code, uint32_t *data)
{
	char control[255], buf[8];
	struct msghdr msg = {.msg_iov = NULL};
	struct cmsghdr *cmsg;
	struct iovec iov;
	int ret = 0;

	*fd = -1;
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	memset(buf, 0, sizeof(buf));
	ret = recvmsg(afs_socket, &msg, 0);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (iov.iov_len != sizeof(buf))
		return -E_AFS_SHORT_READ;
	*code = *(uint32_t*)buf;
	*data =  *(uint32_t*)(buf + 4);
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_SOCKET
			|| cmsg->cmsg_type != SCM_RIGHTS)
		continue;
		if ((cmsg->cmsg_len - CMSG_LEN(0)) / sizeof(int) != 1)
		continue;
		*fd = *(int *)CMSG_DATA(cmsg);
	}
	return 1;
}

static void recv_afs_result(struct vss_task *vsst)
{
	int ret, passed_fd, shmid;
	uint32_t afs_code = 0, afs_data = 0;
	struct stat statbuf;

	vsst->afsss = AFS_SOCKET_READY;
	mmd->afd.afhi.chunk_table = NULL;
	ret = recv_afs_msg(vsst->afs_socket, &passed_fd, &afs_code, &afs_data);
	if (ret < 0)
		goto err;
	PARA_DEBUG_LOG("fd: %d, code: %u, shmid: %u\n", passed_fd, afs_code,
		afs_data);
	ret = -E_NOFD;
	if (afs_code != NEXT_AUDIO_FILE)
		goto err;
	if (passed_fd < 0)
		goto err;
	shmid = afs_data;
	ret = load_afd(shmid, &mmd->afd);
	if (ret < 0)
		goto err;
	shm_destroy(shmid);
	ret = fstat(passed_fd, &statbuf);
	if (ret < 0) {
		PARA_ERROR_LOG("fstat error:\n");
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	mmd->size = statbuf.st_size;
	mmd->mtime = statbuf.st_mtime;
	vsst->map = para_mmap(mmd->size, PROT_READ, MAP_PRIVATE,
		passed_fd, 0);
	close(passed_fd);
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	mmd->events++;
	mmd->num_played++;
	mmd->new_vss_status_flags &= (~VSS_NEXT);
	tv_add(now, &vsst->announce_tv, &vsst->data_send_barrier);
	afh_get_header(&mmd->afd.afhi, vsst->map, &vsst->header_buf,
		&vsst->header_len);
	return;
err:
	free(mmd->afd.afhi.chunk_table);
	if (passed_fd >= 0)
		close(passed_fd);
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	mmd->new_vss_status_flags = VSS_NEXT;
}

/**
 * Main sending function.
 *
 * This function gets called from para_server as soon as the next chunk of data
 * should be pushed out. It obtains a pointer to the data to be sent out as
 * well as its length from mmd->afd.afhi. This information is then passed to
 * each supported sender's send() function which is supposed to send out the data
 * to all connected clients.
 */
static void vss_send_chunk(struct vss_task *vsst)
{
	int i;
	struct timeval due;
	const char *buf;
	size_t len;

	if (!vsst->map || !vss_playing())
		return;
	compute_chunk_time(mmd->chunks_sent, &mmd->afd.afhi.chunk_tv,
		&mmd->stream_start, &due);
	if (tv_diff(&due, now, NULL) > 0)
		return;
	if (chk_barrier("eof", &vsst->eof_barrier, &due, 1) < 0)
		return;
	if (chk_barrier("data send", &vsst->data_send_barrier,
			&due, 1) < 0)
		return;
	if (mmd->current_chunk >= mmd->afd.afhi.chunks_total) { /* eof */
		mmd->new_vss_status_flags |= VSS_NEXT;
		return;
	}
	/*
	 * We call the send function also in case of empty chunks as they
	 * might have still some data queued which can be sent in this case.
	 */
	if (!mmd->chunks_sent) {
		struct timeval tmp;
		mmd->stream_start = *now;
		tv_scale(mmd->current_chunk, &mmd->afd.afhi.chunk_tv, &tmp);
		mmd->offset = tv2ms(&tmp);
		mmd->events++;
	}
	afh_get_chunk(mmd->current_chunk, &mmd->afd.afhi, vsst->map, &buf, &len);
	for (i = 0; senders[i].name; i++)
		senders[i].send(mmd->current_chunk, mmd->chunks_sent, buf, len,
			vsst->header_buf, vsst->header_len);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->chunks_sent++;
	mmd->current_chunk++;
}

static void vss_post_select(struct sched *s, struct task *t)
{
	int ret, i;
	struct vss_task *vsst = container_of(t, struct vss_task, task);

	if (mmd->sender_cmd_data.cmd_num >= 0) {
		int num = mmd->sender_cmd_data.cmd_num,
			sender_num = mmd->sender_cmd_data.sender_num;

		if (senders[sender_num].client_cmds[num])
			senders[sender_num].client_cmds[num](&mmd->sender_cmd_data);
		mmd->sender_cmd_data.cmd_num = -1;
	}
	if (vsst->afsss != AFS_SOCKET_CHECK_FOR_WRITE) {
		if (FD_ISSET(vsst->afs_socket, &s->rfds))
			recv_afs_result(vsst);
	} else if (FD_ISSET(vsst->afs_socket, &s->wfds)) {
		PARA_NOTICE_LOG("requesting new fd from afs\n");
		ret = send_buffer(vsst->afs_socket, "new");
		vsst->afsss = AFS_SOCKET_AFD_PENDING;
	}
	for (i = 0; senders[i].name; i++) {
		if (!senders[i].post_select)
			continue;
		senders[i].post_select(&s->rfds, &s->wfds);
	}
	vss_send_chunk(vsst);
}

/**
 * Initialize the virtual streaming system task.
 *
 * \param afs_socket The fd for communication with afs.
 *
 * This also initializes all supported senders and starts streaming
 * if the --autoplay command line flag was given.
 */
void init_vss_task(int afs_socket)
{
	static struct vss_task vss_task_struct, *vsst = &vss_task_struct;
	int i;
	char *hn = para_hostname(), *home = para_homedir();
	long unsigned announce_time = conf.announce_time_arg > 0?
			conf.announce_time_arg : 300,
		autoplay_delay = conf.autoplay_delay_arg > 0?
			conf.autoplay_delay_arg : 0;

	vsst->afs_socket = afs_socket;
	vsst->task.pre_select = vss_pre_select;
	vsst->task.post_select = vss_post_select;
	ms2tv(announce_time, &vsst->announce_tv);
	PARA_INFO_LOG("announce timeval: %lums\n", tv2ms(&vsst->announce_tv));
	for (i = 0; senders[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s sender\n", senders[i].name);
		senders[i].init(&senders[i]);
	}
	free(hn);
	free(home);
	mmd->sender_cmd_data.cmd_num = -1;
	make_empty_status_items(mmd->afd.verbose_ls_output);
	if (conf.autoplay_given) {
		struct timeval tmp;
		mmd->vss_status_flags |= VSS_PLAYING;
		mmd->new_vss_status_flags |= VSS_PLAYING;
		ms2tv(autoplay_delay, &tmp);
		tv_add(now, &tmp, &vsst->autoplay_barrier);
	}
	register_task(&vsst->task);
}
